#! /usr/bin/env python3

# NEED log sorted by default, and only unique run parameters
# NEED percent positive mode
# NEED to try MCL using the binary
# NEED to make rerunning aggclus not use old data

import os
import re
import sys
import copy
import shutil
import string
import random
import tempfile
import argparse
import itertools
import subprocess
import pandas as pd
from collections import defaultdict
from mycotools.lib.kontools import multisub, findExecs, format_path, \
    eprint, vprint, read_json, write_json, mkOutput, fmt_float
from mycotools.lib.biotools import fa2dict, dict2fa
sys.setrecursionlimit(1000000)

class ClusteringError(Exception):
    pass

class ClusterParameterError(Exception):
    pass

def run_mmseqs(fa_path, res_base, wrk_dir, algorithm = 'mmseqs easy-linclust',
                 min_id = 0.3, clus_const = 0.5, cpus = 1, verbose = False):
    res_path = res_base + '_cluster.tsv'
    tmp_dir = os.path.dirname(res_path) + '/tmp/'
    if verbose:
        stdout, stderr = None, None
    else:
        stdout = subprocess.DEVNULL
        stderr = subprocess.DEVNULL
    mmseqs_cmd = algorithm.split(' ')
    mmseqs_cmd.extend([fa_path, res_base, tmp_dir,
                       '--min-seq-id', fmt_float(min_id), '--threads',
                       str(cpus), '--compressed', '1',
                       '--cov-mode', '0', '-c', str(clus_const),
                       '-e', '0.1', '-s', '7.5'])
    mmseqs_exit = subprocess.call(mmseqs_cmd,
                                  stdout = stdout,
                                  stderr = stderr)
    if mmseqs_exit:
        raise ClusteringError('Clustering failed: ' + str(mmseqs_exit) \
                            + ' ' + str(mmseqs_cmd))
    if os.path.isfile(res_base + '_all_seqs.fasta'):
        os.remove(res_base + '_all_seqs.fasta')
    if os.path.isfile(res_base + '_rep_seq.fasta'):
        os.remove(res_base + '_rep_seq.fasta')
    if os.path.isdir(tmp_dir):
        shutil.rmtree(tmp_dir)
    return res_path

def parse_mmseqs_clus(res_path):
    derivations = defaultdict(list)
    with open(res_path, 'r') as raw:
        for line in raw:
            k, v = line.rstrip().split() # all white space
            derivations[k].append(v)

    hg_list = []
    for k, v in derivations.items():
        v.append(k)
        hg_list.append(sorted(set(v)))

    hg2gene = {i: v for i, v in enumerate(
                    sorted(hg_list, key = lambda x: len(x),
                    reverse = True)
                    )}
    gene2hg = {}
    for hg, genes in hg2gene.items():
        for gene in genes:
            gene2hg[gene] = hg
    return hg2gene, gene2hg

def makeDmndDB(diamond, queryFile, output_dir, cpus = 1):
    """create a diamond database:
    diamond: binary path, queryFile: query_path"""
    outputDB = output_dir + re.sub(r'\.[^\.]+$', '', os.path.basename(queryFile))
    cmd = [
        diamond, 'makedb', '--in', queryFile,
        '--db', outputDB, '--threads', str(cpus)
        ]
    dmndDBcode = subprocess.call(
        cmd, stdout = subprocess.DEVNULL, stderr = subprocess.DEVNULL
        )
    return outputDB, dmndDBcode

def runDmnd(
    diamond, queryFile, queryDB, outputFile, pid = True,
    cpus = 1, verbose = False, blast = 'blastp', minid = 15
    ):
    """run diamond, no self-hits, more-sensitive"""

    if pid:
        distanceVal = 'pident'
    else:
        distanceVal = 'bitscore'

    cmd = [
        diamond, blast, '-d', queryDB, '-q', queryFile,
        '--out', outputFile, '--threads', str(cpus),
        '--outfmt', '6', 'qseqid', 'sseqid', distanceVal,
        '--id', str(minid), '--no-self-hits', '--more-sensitive'
        ]

    if not verbose:
        dmndCode = subprocess.call(
            cmd, stdout = subprocess.DEVNULL, stderr = subprocess.DEVNULL
            )
    else:
        dmndCode = subprocess.call(
            cmd
            )
    return outputFile, dmndCode

def rd_dmnd_distmtx( outputFile, minVal, pid = True ):
    """Read a diamond output as a distance matrix"""

    dist_dict = {}
    with open(outputFile, 'r') as raw:
        if pid:
            for line in raw:
                q, s, v0 = line.rstrip().split('\t')
                v = round(float(v0))
                if q == s:
                    continue
                if v > minVal:
                    if q not in dist_dict:
                        dist_dict[q] = {}
                        dist_dict[q][q] = 0.0
                    if s not in dist_dict:
                        dist_dict[s] = {}
                        dist_dict[s][s] = 0.0
                    dist_dict[q][s] = 100 - v
                    dist_dict[s][q] = 100 - v
            distance_matrix = pd.DataFrame(dist_dict).sort_index(1).sort_index(0).fillna(100)
        else:
             for line in raw:
                q, s, v0 = line.rstrip().split('\t')
                v = float(v0)
                if q == s:
                    continue
                if v > minVal:
                    if q not in dist_dict:
                        dist_dict[q] = {}
                        dist_dict[q][q] = 0.0
                    if s not in dist_dict:
                        dist_dict[s] = {}
                        dist_dict[s][s] = 0.0
             distance_matrix = pd.DataFrame(dist_dict).sort_index(1).sort_index(0).fillna(0)


    if pid:
        distance_matrix /= 100
    else:
        minV, maxV = min(distance_matrix), max(distance_matrix)
        denom = maxV - minV
        distance_matrix -= minV
        distance_matrix /= denom
        distance_matrix = 1 - distance_matrix

    return distance_matrix #, outMatrix

def runUsearch( fasta, output, clus_var, cpus = 1, verbose = False ):

    if verbose:
        subprocess.call([
            'usearch', '-calc_distmx', fasta,
            '-tabbedout', output, '-clus_var',
            clus_var, '-threads', str(cpus)
            ] #stdout = subprocess.PIPE,
            #stderr = subprocess.PIPE
            )
    else:
        subprocess.call([
            'usearch', '-calc_distmx', fasta,
            '-tabbedout', output, '-clus_var',
            clus_var, '-threads', str(cpus)
            ], stdout = subprocess.DEVNULL,
            stderr = subprocess.DEVNULL
            )

def rd_usrch_distmtx( dis_path, sep = '\t' ):
    '''Imports a distance matrix with each line formatted as `organism $SEP organism $SEP distance`.
    - this is equivalent to the `-tabbedout` argument in `usearch -calc_distmx`. The function 
    compiles a dictionary with the information and reciprocal information for each organism in each
    line, then converts this dictionary of dictionaries into a pandas dataframe. As a distance matrix,
    NA values are converted to maximum distance (1).'''

    distance_matrix = pd.DataFrame()
    with open( dis_path, 'r' ) as raw:
        data = raw.read()
    prepData = [ x.rstrip() for x in data.split('\n') if x != '' ]

    dist_dict = {}
    for line in prepData:
        vals = line.split( sep )
        if vals[0] not in dist_dict:
            dist_dict[ vals[0] ] = {}
        if vals[1] not in dist_dict:
            dist_dict[ vals[1] ] = {}
        dist_dict[ vals[0] ][ vals[1] ] = float(vals[2])
        dist_dict[ vals[1] ][ vals[0] ] = float(vals[2])

    distance_matrix = pd.DataFrame( dist_dict ).sort_index(0).sort_index(1)

    return distance_matrix.fillna( 1.0 )

def scikitaggd( distance_matrix, maxDist = 0.6, linkage = 'single' ):
    '''Performs agglomerative clustering and extracts the cluster labels, then sorts according to
    cluster number. In the future, this will also extract a Newick tree.'''

    clustering = AgglomerativeClustering( 
        affinity = 'precomputed',
        distance_threshold = maxDist,
        n_clusters = None,
        linkage = linkage
        ).fit( distance_matrix )

    i = iter( distance_matrix.columns )
    clusters = { next( i ): x for x in clustering.labels_ }
    clusters = { k: v for k, v in sorted( clusters.items(), key = lambda item: item[1] ) }

    return clusters

def getNewick(node, newick, parentdist, leaf_names):
    '''Code from https://stackoverflow.com/questions/28222179/save-dendrogram-to-newick-format
    adopted from @jfn'''
    if node.is_leaf():
        return "%s:%.2f%s" % (leaf_names[node.id], parentdist - node.dist, newick)
    else:
        if len(newick) > 0:
            newick = "):%.2f%s" % (parentdist - node.dist, newick)
        else:
            newick = ");"
        newick = getNewick(node.get_left(), newick, node.dist, leaf_names)
        newick = getNewick(node.get_right(), ",%s" % (newick), node.dist, leaf_names)
        newick = "(%s" % (newick)
        return newick

def getClusterLabels( labels, clusters ):

    i = iter(labels)
    protoclusters = {next(i): x for x in clusters}
    clusters = {k: v for k, v in sorted(protoclusters.items(), key = lambda item: item[1])}

    return clusters

def runMCL(distMat, inflation):

    key2gene = {i: v for i, v in enumerate(list(distMat.keys()))}
    npMat = distMat.to_numpy()
    result = mc.run_mcl(npMat, inflation = inflation)
    protoClusters = mc.get_clusters(result)

    clusters = {}
    for i, keys in enumerate(protoClusters):
        clusters = {**clusters, **{key2gene[key]: i for key in keys}}

    return clusters

def scipyaggd( distMat, maxDist, method = 'single' ):
    '''Performs agglomerative clustering using SciPy and extracts the cluster labels, then sorts
    according to cluster number.'''

    squareform_matrix = squareform(distMat.values)
    linkage_matrix = hierarchy.linkage(squareform_matrix, method)
    tree = hierarchy.to_tree(linkage_matrix)
    fcluster = hierarchy.fcluster(
        linkage_matrix, maxDist, 
        criterion = 'distance'
        )
    clusters = getClusterLabels(distMat.index, fcluster)

    return clusters, tree

def dmnd_main(fa_path, minVal, output_dir, distFile, pid = True, verbose = False, cpus = 1):
    queryDB, makeDBcode = makeDmndDB('diamond', fa_path, output_dir, cpus = cpus)
    dmndOut, dmndCode = runDmnd(
        'diamond', fa_path, queryDB, distFile + '.tmp', pid = pid,
        cpus = cpus, verbose = verbose, blast = 'blastp'
        )
    shutil.move(distFile + '.tmp', distFile)
    distance_matrix = rd_dmnd_distmtx( distFile, minVal, pid = pid )
    return distance_matrix
    
def usrch_main(fasta, min_id, output, cpus = 1, verbose = False):
    vprint('\nusearch aligning', flush = True, v = verbose)
    runUsearch(fasta, output + '.dist', str(1 - min_id), cpus, verbose)
    distance_matrix = rd_usrch_distmtx( output + '.dist' )
    return distance_matrix

def readLog(log_path, newLog):
    oldLog = read_json(log_path)
    if oldLog['fasta'] == newLog['fasta'] and \
        oldLog['minimum_id'] == newLog['minimum_id'] and \
        oldLog['search_program'] == newLog['search_program']:
        newLog['iterations'] = oldLog['iterations']
        newLog['successes'] = oldLog['successes']
        for i in newLog['iterations']:
            i['cluster'] = tuple(i['cluster'])
        for i in newLog['successes']:
            i['cluster'] = tuple(i['cluster'])
    elif newLog['distance_matrix']:
        if os.path.isfile(newLog['distance_matrix']):
            os.remove(newLog['distance_matrix'])
    return newLog

def extract_closest_cluster(iterations, min_seq, max_seq):
    for i in iterations:
        if i < min_seq:
            i['discrep'] = min_seq - i['size']
        else:
            i['discrep'] = i['size'] - max_seq
    sorts = sorted(iterations, key = lambda x: x['discrep'])
    return sorts[0]

def sort_iterations(iterations, reverse = True):
    return sorted([dict(t) for t in {
                    tuple(d.items()) for d in iterations
                    }], key = lambda x: (x['size'], x['cluster_variable']),
                    reverse = reverse)

def cluster_iter_mmseqs(params, min_seq, max_seq, clus_const, clus_var,
               interval, min_var, max_var, log_dict, log_path, 
               focal_gene = None, verbose = False, spacer = '\t', 
               cpus = 1):

    oldDirection, attempt = None, 0
    if focal_gene:
        res_base = params['dir'] + 'working/' + focal_gene
    else:
        res_base = params['dir'] + 'working/' \
                 + re.sub(r'\.[^\.]+$', '', params['fa']) 
    while clus_var >= min_var and clus_var <= max_var:
        attempt += 1
        name = '_c' + str(round(clus_const*1000) / 1000) \
             + '_v' + str(round(clus_var*1000) / 1000)
        res_info = res_base + name
        res_path = res_info + '_cluster.tsv'
        if not os.path.isfile(res_path):
            run_mmseqs(params['fa'], res_info, 
                       params['dir'] + 'working/',
                       algorithm = params['bin'],
                       min_id = clus_var, verbose = verbose,
                       clus_const = params['clus_const'], cpus = cpus)
        cluster_dict, clusters = parse_mmseqs_clus(res_path)
        cluster = cluster_dict[clusters[focal_gene]]
        focal_len = len(cluster) 

        vprint('\nITERATION ' + str(attempt) + ': ' + focal_gene \
             + ' cluster size: ' + str(focal_len), flush = True, v = verbose)
        vprint('Cluster parameter: ' + str(clus_var), flush = True, v = verbose)
        iteration_dict = {'size': focal_len, 'cluster_variable': clus_var,
                          'cluster': tuple(cluster)}
        log_dict['iterations'].append(iteration_dict)
        log_dict['iterations'] = sort_iterations(log_dict['iterations'])
        if focal_len >= min_seq: # if greater than minimum sequences
            if max_seq: # if there is a max set of sequences
                if focal_len <= max_seq: # if less than max sequences
                    vprint(spacer + '\tSUCCESS!', flush = True, v = verbose)
                    exit_code = 0
                    direction = -1
                    log_dict['successes'].append(iteration_dict)
                    log_dict['successes'] = \
                        sort_iterations(log_dict['successes'])
                else: # descend
                    direction = 1
            else: # no max sequences, minimum is met, this was successful
                vprint(spacer + '\tSUCCESS!', flush = True, v = verbose)
                exit_code = 0
                log_dict['successes'].append(iteration_dict)
                log_dict['successes'] = sort_iterations(log_dict['successes'])
                break
        else: # need to increase cluster size
            direction = -1

        if oldDirection: # if there is an older iteration
            if direction != oldDirection: # if the direction switched
                diff_len = ofocal_len - focal_len
                if oldDirection * -1 == diff_len/abs(diff_len):
                # percent ID change is congruent with the size change, which should
                # not be the case, i.e. decreasing % ID shouldn't decrease
                # module size
                    pass
                elif log_dict['successes']:
                    # grab the largest success
                    cluster = log_dict['successes'][0]['cluster']
                    exit_code = 0
                    break
                else:
                    eprint(spacer + 'WARNING: Overshot - ' \
                         + 'could not find parameters using current interval', 
                           flush = True)
                    iteration = extract_closest_cluster(log_dict['iterations'],
                                                        min_seq, max_seq)
                    cluster = iteration['cluster']
                    exit_code = 2
                    break
        else: # set the old iteration direction detail
            oldDirection = direction
            exit_code = 1
#         clus_const += (interval*direction)

        ofocal_len = copy.copy(focal_len)
        oclus_var = copy.copy(clus_var)
        clus_var += (interval*direction)

        # clus_var should be to 3 significant figures
        clus_var = round(clus_var * 100)/100
        # adjust the clus_vareter based on the direction it should change
        if oclus_var != min_var and oclus_var != max_var:
            if clus_var < min_var: # set as minimum value if it descends below
                clus_var = min_var
            elif clus_var > max_var: # set as maximum if it ascends above
                clus_var = max_var

    write_json(log_dict, log_path)
    try:
        return cluster, log_dict, exit_code
    except UnboundLocalError: # cluster not defined, didn't enter while loop
        raise ClusterParameterError('failed to run with given parameters')

def cluster_iter_aggclus(params, min_seq, max_seq, clus_const, clus_var,
                interval, min_var, max_var, log_dict, log_path, 
                focal_gene = None, verbose = False, spacer = '\t', 
                cpus = 1):

    oldDirection, attempt = None, 0
    if focal_gene:
        res_base = params['dir'] + 'working/' + focal_gene
    else:
        res_base = params['dir'] + 'working/' \
                 + re.sub(r'\.[^\.]+$', '', params['fa']) 
    while clus_var >= min_var and clus_var <= max_var:
        attempt += 1
        if 100 - round(100*clus_var) < round(100*clus_const):
            break
        name = '_c' + str(round(clus_const*1000) / 1000) \
             + '_v' + str(round(clus_var*1000) / 1000)
        clusters, tree = scipyaggd(params['dist'],
                                   float(clus_var), 
                                   params['link']) # cluster
        newick = getNewick(tree, "", tree.dist, list(params['dist'].index))
        write_data(newick, clusters, res_base + name)

        cluster_dict = defaultdict(list) # could be more efficient by grabbing in getClusterLabels
        for gene, index in clusters.items(): # create a dictionary clusID:
            # [genes]
            cluster_dict[index].append(gene)

        try:
            cluster = cluster_dict[clusters[focal_gene]]
            focal_len = len(cluster) 
        except KeyError: # focal gene not in output
            focal_len = 0
            cluster = None
            newick = ''


        vprint('\nITERATION ' + str(attempt) + ': ' + focal_gene \
             + ' cluster size: ' + str(focal_len), flush = True, v = verbose)
        vprint('Cluster parameter: ' + str(clus_var), flush = True, v = verbose)
        iteration_dict = {'size': focal_len, 'cluster_variable': clus_var,
                          'cluster': tuple(cluster), 'tree': newick}
        log_dict['iterations'].append(iteration_dict)
        log_dict['iterations'] = sort_iterations(log_dict['iterations'])
        if focal_len >= min_seq: # if greater than minimum sequences
            if max_seq: # if there is a max set of sequences
                if focal_len <= max_seq: # if less than max sequences
                    vprint(spacer + '\tSUCCESS!', flush = True, v = verbose)
                    exit_code = 0
                    direction = -1
                    log_dict['successes'].append(iteration_dict)
                    log_dict['successes'] = \
                        sort_iterations(log_dict['successes'])
                else: # descend
                    direction = 1
            else: # no max sequences, minimum is met, this was successful
                vprint(spacer + '\tSUCCESS!', flush = True, v = verbose)
                exit_code = 0
                log_dict['successes'].append(iteration_dict)
                log_dict['successes'] = \
                    sort_iterations(log_dict['successes'])
                break
        else: # need to increase cluster size
            exit_code = 1
            direction = -1

        if oldDirection: # if there is an older iteration
            if direction != oldDirection: # if the direction switched
                if log_dict['successes']:
                    # sort by the largest success
                    cluster = log_dict['successes'][0]['cluster']
                    newick = log_dict['successes'][0]['tree']
                    exit_code = 0
                else:
                    eprint(spacer + 'WARNING: Overshot - ' \
                         + 'could not find parameters using current interval', 
                           flush = True)
                    iteration = extract_closest_cluster(log_dict['iterations'],
                                                        min_seq, max_seq)
                    cluster = iteration['cluster']
                    newick = iteration['tree']
                    exit_code = 2
                break
        else: # set the old iteration direction detail
            oldDirection = direction
#         clus_const += (interval*direction)

        oclus_var = copy.copy(clus_var)
        clus_var -= (interval*direction)

        # clus_var should be to 3 significant figures
        clus_var = round(clus_var * 100)/100
        # adjust the clus_vareter based on the direction it should change
        if oclus_var != min_var and oclus_var != max_var:
            if clus_var < min_var: # set as minimum value if it descends below
                clus_var = min_var
            elif clus_var > max_var: # set as maximum if it ascends above
                clus_var = max_var

    write_json(log_dict, log_path)
    try:
        return cluster, newick, log_dict, exit_code
    except UnboundLocalError: # cluster not defined, didn't enter while loop
        raise ClusterParameterError('failed to run with given parameters')


def main(
    fa_path, clus_const, clus_var, min_seq, max_seq, 
    search_program = 'diamond', linkage = 'single',
    focal_gene = False, interval = 0.1, output= None, cpus = 1,
    verbose = False, spacer = '\t\t', log_path = None,
    pid = True, dmnd_dir = None, min_var = 0.05,
    max_var = 1
    ):

    if not os.path.isdir(os.path.dirname(output) + '/working/'):
        os.mkdir(os.path.dirname(output) + '/working/')

    if search_program in {'usearch', 'diamond'}:
        algorithm = 'hierarchical'
        dist_path = output + '.dist'
    else:
        algorithm = search_program
        dist_path = None

    overshot = False
    log_dict = {
        'algorithm': algorithm, 'fasta': fa_path, 
        'minimum_id': clus_const, 'cluster_variable': clus_var,
        'minimum_sequences': min_seq, 'maximum_sequences': max_seq, 
        'search_program': search_program, 'distance_matrix': dist_path, 
        'linkage': linkage, 'focal_gene': focal_gene,
        'iterations': [], 'successes': []
        }
    if log_path:
        if os.path.isfile(log_path):
            log_dict = readLog(log_path, log_dict)
        write_json(log_dict, log_path)

    if algorithm == 'hierarchical':
        global hierarchy
        global squareform
        from scipy.cluster import hierarchy
        from scipy.spatial.distance import squareform
        param_dict = {'dist': None, 'link': linkage, 'dir': os.path.dirname(output) + '/'}
        if os.path.isfile(log_dict['distance_matrix']):
            if search_program == 'usearch':
                param_dict['dist'] = rd_usrch_distmtx(log_dict['distance_matrix'])
            else:
                param_dict['dist'] = rd_dmnd_distmtx(log_dict['distance_matrix'], clus_const, pid = pid)
        else:
            if search_program == 'diamond': #elif if above lines not highlighted
                if not dmnd_dir:
                    dmnd_dir = os.path.dirname(log_dict['distance_matrix']) + '/'
                param_dict['dist'] = dmnd_main(
                    fa_path, clus_const, dmnd_dir, 
                    log_dict['distance_matrix'], pid = pid, 
                    verbose = verbose, cpus = cpus
                    )
            elif search_program == 'usearch':
                param_dict['dist'] = usrch_main(fa_path, clus_const, output, cpus, verbose)
    else:
        param_dict = {'fa': fa_path, 'bin': search_program, 
                      'clus_const': clus_const, 
                      'dir': os.path.dirname(output) + '/'}

    if focal_gene:
        vprint('\nClustering', flush = True, v = verbose)
        if algorithm == 'hierarchical':
            cluster, newick, log_dict, error = cluster_iter_aggclus(
                   param_dict, min_seq, max_seq, clus_const, clus_var,
                   interval, min_var, max_var, log_dict, log_path, 
                   focal_gene, verbose, spacer, cpus
                   )
            if error == 2:
                overshot = True
            else:
                overshot = False
            return cluster, newick, overshot, log_dict
        else:
            cluster, log_dict, error = cluster_iter_mmseqs(
                   param_dict, min_seq, max_seq, clus_const, clus_var,
                   interval, min_var, max_var, log_dict, log_path, 
                   focal_gene, verbose, spacer, cpus
                   )
            if error == 2:
                overshot = True
            else:
                overshot = False
            return cluster, None, overshot, log_dict
    else:
        if focal_gene:
            res_base = param_dict['dir'] + focal_gene
        else:
            res_base = param_dict['dir'] \
                     + re.sub(r'\.[^\.]+$', '', os.path.basename(fa_path))
        if algorithm == 'hierarchical':
            clusters, tree = scipyaggd(param_dict['dist'],
                                       float(clus_var), 
                                       param_dict['link']) # cluster
            newick = getNewick(tree, "", tree.dist, 
                               list(param_dict['dist'].index))
            write_data(newick, clusters, res_base)
            return clusters, None, None, log_dict
        else:
            res_path = run_mmseqs(param_dict['fa'], res_base, 
                                    param_dict['dir'] + 'working/',
                                    algorithm = param_dict['bin'],
                                    min_id = clus_var, verbose = verbose,
                                    clus_const = param_dict['clus_const'], cpus = cpus)
            cluster_dict, clusters = parse_mmseqs_clus(res_path)
            write_data(None, clusters, param_dict['dir'], res_base)
            return clusters, None, None, log_dict
    
def write_data(newick, clusters, output):

    #clusters = scikitaggd( distance_matrix, float(args.max_dist), args.linkage )
    out_str = '\n'.join([i + '\t' + str(clusters[i]) for i in clusters])
    with open(output + '.clus', 'w') as out:
        out.write(out_str)
    if newick:
        with open(output + '.newick', 'w') as out:
            out.write(newick)


if __name__ == '__main__':

    parser = argparse.ArgumentParser( 
        description = "Sequence clustering with iterative search option." 
        )
    parser.add_argument('-f', '--fasta', required = True)
    parser.add_argument( 
        '-a', '--alignment', 
        help = 'Alignment software {"mmseqs", "diamond", "usearch"}; ' \
             + 'diamond and usearch call hierarchical clustering. ' \
             + 'DEFAULT: mmseqs', 
        default = 'mmseqs' 
        )
    parser.add_argument('-l', '--linclust', action = 'store_true',
        help = 'Use linclust instead of mmseqs cluster (faster, less sensitive)')
    parser.add_argument('-m', '--cluster_constant', 
        help = 'DEFAULT: [mmseqs]: 0.4 percent query coverage, '  \
             + '[aggclus]: 0.2 percent identity, [aggclus]: 30 bitscore', 
        type = float)
    parser.add_argument('-x', '--cluster_variable', 
        help = '[mmseqs]: Minimum identity; DEFAULT: 0.3; ' \
            +  '[aggclus]: Maximum distance; DEFAULT: 0.75', 
            type = float)
    parser.add_argument('-i', '--iterative', 
        help = '[--min_seq]: Gene to iteratively cluster for; '  \
            + 'eliminates -x')
    parser.add_argument('--min_seq', type = int, help = '[-i]: Iteratively adjust -m and -x until a cluster ' \
        + ' of minimum sequences is obtained; DEFAULT: 2', default = 2)
    parser.add_argument('--max_seq', type = int, 
        help = '[-i]: Maximum sequences for iterative clustering; ' \
            + 'DEFAULT 100', default = 100)
#    parser.add_argument('-r', '--refine', action = 'store_true', 
 #       help = '[--max_seq]: Refine cluster to maximize sequences within parameters.')
#    parser.add_argument('-i', '--inflation', type = float, default = 1.5,
 #       help = 'Inflation value for MCL; DEFAULT: 1.5')
    parser.add_argument(
        '-d', '--distance_type', help = '[aggclus]: {"identity", "bitscore"} ' + \
        'bitscore is normalized 0-1 following -m and -x should be 0-1; DEFAULT: identity',
        default = 'identity'
        )

    parser.add_argument('--linkage', default = 'single', 
        help = "[aggclus]: Linkage criterion: " \
        + "'complete' (maximum distance), 'average', 'weighted', " \
        + "'centroid', or 'single' (minimum distance) DEFAULT: 'single' (minimum)")
    parser.add_argument('-o', '--output')
    parser.add_argument('-c', '--cpus', default = 1, type = int)
    parser.add_argument('-v', '--verbose', action = 'store_true')
    args = parser.parse_args()

#    if args.refine and not args.max_seq:
 #       eprint('\nERROR: --max_seq required for refinement', flush = True)
    if args.alignment in {'diamond', 'usearch'}:
        if args.linkage not in {'complete', 'average', 'weighted', 'centroid', 'single'}:
            eprint('\nERROR: Invalid linkage criterium', flush = True)
            sys.exit(1)
        elif args.distance_type not in {'identity', 'bitscore'}:
            eprint('\nERROR: Invalid distance type', flush = True)
            sys.exit(3)
        if args.distance_type == 'identity':
            pid = True
            if not args.cluster_constant:
                args.cluster_constant = 0.2
        else:
            pid = False
            if not args.cluster_constant:
                args.cluster_constant = 30
        if not args.cluster_variable:
            args.cluster_variable = 0.75
    elif args.alignment == 'mmseqs':
        if args.linclust:
            args.alignment = 'mmseqs easy-linclust'
        else:
            args.alignment = 'mmseqs easy-cluster'
        pid = None
        if not args.cluster_constant:
            args.cluster_constant = 0.4
        if not args.cluster_variable:
            args.cluster_variable = 0.3
    else:
        eprint('\nERROR: Invalid alignment software', flush = True)
        sys.exit(2)

    findExecs([args.alignment.split()[0]], 
               exit = set(args.alignment.split()[0]))
    interval = 0.1

#    if args.iterative and args.alignment in {'diamond', 'usearch'}:
 #       clus_var = 1 - args.cluster_constant
    if 1 - args.cluster_variable <= args.cluster_constant \
        and args.alignment != 'mmseqs':
        eprint('\nWARNING: 1 - maximum distance exceeds minimum connection, clustering is ineffective', flush = True)
        sys.exit(3)
    else:
        clus_var = args.cluster_variable
 #   else:
#        clus_var = args.inflation

    fa_path = format_path(args.fasta)
    fa = fa2dict(fa_path)
    if args.iterative:
        if len(fa) < args.min_seq:
            eprint('\nERROR: minimum sequences is greater than fasta input', flush = True)
            sys.exit(5)
        elif args.iterative not in fa:
            eprint('\nERROR: ' + args.iterative + ' not in ' + fa_path, flush = True)
            sys.exit(6)

    if args.output:
        if not os.path.isdir(format_path(args.output)):
            os.mkdir(format_path(args.output))
        dmnd_dir = format_path(args.output)
        output = dmnd_dir \
            + re.sub(r'\.[^\.]+$', '', os.path.basename(fa_path))
    else:
        dmnd_dir = mkOutput(os.getcwd() + '/', 'fa2clus')
        output = dmnd_dir \
            + re.sub(r'\.[^\.]+$', '', os.path.basename(fa_path))

    cluster, tree, overshot, log_dict = main(
        fa_path, args.cluster_constant, clus_var, args.min_seq, args.max_seq, 
        search_program = args.alignment, linkage = args.linkage, verbose = args.verbose,
        focal_gene = args.iterative, interval = interval, output = output,
        log_path = dmnd_dir + '.' + os.path.basename(fa_path) + '.fa2clus.json',
        pid = pid, dmnd_dir = dmnd_dir, cpus = args.cpus,
        spacer = '\n'
        )

    if tree:
        with open(output + '.newick', 'w') as out:
            out.write(tree)
    if args.iterative:
        input_fa = fa2dict(fa_path)
        try:
            output_fa = {x: input_fa[x] for x in cluster}
        except TypeError:
            eprint('\nERROR: empty cluster', flush = True)
            sys.exit(10)
        with open(output + '.fa', 'w') as out:
            out.write(dict2fa(output_fa)) 

    sys.exit(0)



#! /usr/bin/env python3

# NEED option to fail upon any failures
# NEED log
# NEED nhmmer option
# NEED to .tmp and move files
# NEED to add alignment extraction only option

import os
import re
import sys
import argparse
import subprocess
import multiprocessing as mp
from io import StringIO
from collections import defaultdict
from mycotools.lib.kontools import intro, outro, collect_files, multisub, \
    findExecs, untardir, eprint, format_path, mkOutput, tardir
from mycotools.lib.dbtools import masterDB, mtdb
from mycotools.lib.biotools import dict2fa
#from mycotools.extractHmmsearch import main as exHmm
from mycotools.acc2fa import dbmain as acc2fa
from mycotools.utils.extractHmmsearch import main as exHmm
from mycotools.utils.extractHmmAcc import grabAccs, main as absHmm

def compile_hmm_cmd(db, hmm_path, output, ome_set = set(), cpu = 1):
    """
    Inputs: mycotools db, hmm_path, output directory, set of omes to ignore.
    Outputs: tuples of arguments for hmmsearches
    """

    cmd_tuples = [ ]
    for ome, row in db.items():
        if ome not in ome_set:
            output_path = output + ome + '.out'
            cmd = (('hmmsearch', '-o', output_path + '.tmp', 
                   '--cpu', str(cpu),
                   hmm_path, row['faa'],
                   '&&',), ('mv', output_path + '.tmp', output_path,),) 
            cmd_tuples.append(cmd)

    return cmd_tuples

def compileextractHmmCmd(db, args, output):
    """
    Inputs: mycotools db, argparse arguments, and output path
    Outputs: tuples of arguments for `run_ex_hmm`
    """

    cmd_tuples = []
    for ome, row in db.items():
        hmmsearch_out = output + ome + '.out'
        cmd_tuples.append((args, hmmsearch_out, output,))

    return tuple(cmd_tuples)


def run_ex_hmm(args, hmmsearch_out, output):

    ome = os.path.basename(hmmsearch_out).replace('.out', '')

    try:
        with open(hmmsearch_out, 'r') as raw:
            data = raw.read()
    except FileNotFoundError:
        eprint('\tWARNING: ' + ome + ' failed', flush = True)
        return ome, False
    out_dict = None
    if len(data) > 100: # check for data # check for data
        out_dict = exHmm(data, args[0], args[1], args[2], args[3], header = False)
        out_dict = {x: [ome, out_dict[x][1]] for x in out_dict}
        # out_dict = {query: [ome, alignment]}
        # does this overwrite other hits?
        return ome, out_dict
    else:
        eprint('\tWARNING: ' + ome + ' empty results', flush = True)
        return ome, False

    

def compileacc2fa(db, output, q_dict):

    cmd_tuples = []
    if not os.path.isdir(output):
        os.mkdir(output)
    fa_dict = {ome: row['faa'] for ome, row in db.items()}
    for q in q_dict:
        cmd_tuples.append((db, q_dict[q], output + '/' + q + '.faa'))

    return cmd_tuples


def runacc2fa(db, q_files, output):

  #  fa_str = ''
 #   for ome in q_files:
#        q = q_files[ome]
#        df = pd.read_csv(StringIO(q), sep = '\t', header = None)
    accs = []
    for ome, hit_data in q_files.items():
        t_accs = [x.split('\t')[0] for x in hit_data.rstrip().split('\n')]
        accs.extend(t_accs)

    fa_str = dict2fa(acc2fa(db, accs))
#    fa_str = acc2fa(db, 
 #       fa_str += dict2fa(acc2fa(df, header, fa = fa_dict[ome])) + '\n'
    
    with open( output, 'w' ) as out:
        out.write( fa_str )


def compile_hmmalign_cmds( output, accessions ):

#    cmd_tuples = [ [], [] ]
    cmds = []
    for acc in accessions:
        fa = output + '/fastas/' + acc + '.faa'
        hmm = output + '/hmms/' + acc + '.hmm'
        align = output + '/aligns/' + acc + '.a2m'
        conv = output + '/aligns/' + acc + '.phylip'
        trim = output + '/trimmed/' + acc + '.trimal'
        if os.path.isfile(conv):
            with open(conv, 'r') as raw:
                data = raw.read()
            if len(data) > 10:
                continue
        cmds.append((('hmmalign', '--outformat', 'A2m', '-o', 
                     align, hmm, fa, '&&',),
                    ('esl-reformat', '-o', conv, 'phylip', align),),)
   #     hmmalign = ( 'hmmalign', '--outformat', 'A2m', '-o', align, hmm, fa )
 #       esl = ( 'esl-reformat', '-o', conv, 'phylip', align )
#        cmd_tuples[0].append( hmmalign )
  #      cmd_tuples[1].append( esl )

    return tuple(cmds)


def compile_trim_cmd(output, mod = '', trimmed = None):

    mod_args = mod.split( ' ' )
    cmd_tuples = []
    aligns = collect_files( output + 'aligns/', 'phylip' )
    if trimmed:
        trimmed = set( os.path.basename(x).replace('.clipkit','') \
            for x in trimmed )
        aligns = [x for x in aligns if os.path.basename(x).replace('.phylip','') \
             not in trimmed]
    for align in aligns:
        trim = output + 'trimmed/' \
             + os.path.basename(align).replace('.phylip','.clipkit') \
             + '.phylip'
        args = ['clipkit', align, '-o', trim]
        if mod_args[0]:
            args.extend(mod_args) 
        cmd_tuples.append( tuple( args ) )

    return cmd_tuples


def main(db, hmm_path, output, accessions, max_hits, query_cov,
         evalue = 0.01, binary = 'hmmsearch', trim_mod = '', verbose = False):

    ome_dir = output + 'omes/'
    aln_dir = output + 'aligns/'
    faa_dir = output + 'fastas/'
    hmm_dir = output + 'hmms/'
    trm_dir = output + 'trimmed/'

    db = db.set_index('ome')
    ome_set, skip1, skip2 = set(), False, False
    if os.path.isdir(faa_dir): # is there a previous run?
        print('\nCompiling previous run', flush = True)
        # check if all fastas are made
        fas = collect_files(faa_dir, '.faa') # grab completed fastas
        ranQueries = [os.path.basename(fa).replace('.faa','') for fa in fas]
        with open(hmm_path, 'r') as hmm_raw:
            queries = grabAccs(hmm_raw.read())
        if not set(queries).difference(set(ranQueries)):
            skip1 = True

# was a fail safe that can be avoided with tmp files
#            for fa in fas:
 #               with open(fa, 'r') as raw:
  #                  data = raw.read()
   #             if len(data) < 10:
    #                skip1 = False
     #               break
        if skip1:
            if os.path.isdir(aln_dir):
                # check if all alignments are made
                phylips = collect_files(aln_dir, 'phylip')
                ranQueries = [os.path.basename(x).replace('.phylip','') \
                              for x in phylips]
                if len(set(queries).intersection(set(ranQueries) \
                    )) == len(set(ranQueries)):
                    skip2 = True
#                    for phy in phylips:
 #                       with open(phy, 'r') as raw:
  #                          data = raw.read()
   #                     if len(data) < 10:
    #                        skip2 = False
     #                       break
            
    if not skip1: # do not skip the first step
        if os.path.isfile(output + 'omes.tar.gz'):
            if not os.path.isdir(ome_dir):
                untardir(output + 'omes.tar.gz')
        # check what reports have been generated
        omes = collect_files(ome_dir, 'out')
        ome_set = set(os.path.basename(x).replace('.out', '') for x in omes)
    else:
        print('\thmmsearch -> hits.faa DONE', flush = True)
        if skip2:
            print('\thits.faa -> phylip DONE', flush = True)

    if not skip1:
        print('\nRunning ' + os.path.basename(binary), flush = True)
        if not os.path.isdir(ome_dir):
            os.mkdir(ome_dir)

        # run hmmer
        par_runs = round(((cpu - 1)/2) - 0.5)
        if par_runs < 1:
            par_cpus = 1
            par_runs = 1
        else:
            par_cpus = 2
        hmmsearch_tuples = compile_hmm_cmd(db, hmm_path, 
                                         ome_dir, 
                                         ome_set = ome_set,
                                         cpu = par_cpus)
        hmmsearch_codes = multisub(hmmsearch_tuples, processes = par_runs,
                                   verbose = verbose)
        for code in hmmsearch_codes:
            if code['code']:
                eprint('\tERROR: ' \
                        + str(code['stdin']),
                        flush = True)

        # extract results
        print('\nExtracting hmmsearch output', flush = True)
        exHmm_args = [accessions, max_hits, query_cov, evalue]
        exHmm_tuples = compileextractHmmCmd(db, exHmm_args, ome_dir)
        with mp.get_context('spawn').Pool(processes = cpu) as pool:
            hmmAligns = pool.starmap(run_ex_hmm, exHmm_tuples)

        mp.Process(target=tardir, args=[ome_dir])
        print( '\nCompiling fastas' , flush = True)
        # q_dict = {query: ome: alignment}
        q_dict = defaultdict(dict)
        for ome, hit in hmmAligns:
            if hit:
                for query in hit:
                    align = hit[query][1]
                    q_dict[query][ome] = align

        acc2fa_tuples = compileacc2fa(db, faa_dir, q_dict)
        with mp.get_context('spawn').Pool(processes = cpu) as pool:
            pool.starmap(runacc2fa, acc2fa_tuples)        

    if args.align:
        if not skip2:
            print( '\nAligning fastas to hmms' , flush = True)
            if not os.path.isdir(hmm_dir):
                os.mkdir(hmm_dir)
            with open(hmm_path, 'r' ) as hmm:
                hmm_dict = absHmm( hmm.read() )
            for accession in hmm_dict:
                with open( hmm_dir + accession + '.hmm', 'w' ) as out:
                    out.write( hmm_dict[ accession ] )

            if not os.path.isdir(aln_dir):
                os.mkdir(aln_dir)

            
            hmmAlign_tuples = compile_hmmalign_cmds(output,
                                                    list(hmm_dict.keys()))
            hmmAlign_codes = multisub(hmmAlign_tuples, processes = cpu - 1,
                                      verbose = verbose)
#            print(hmmAlign_codes)
 #           esl_codes = multisub( hmmAlign_tuples[1], processes = cpu - 1 )
  #          print(esl_codes)
            for index in range(len(hmmAlign_codes)):
                code = hmmAlign_codes[index]
                if code['code']:
                    eprint('\tERROR: ' + str(code['stdin']), flush = True)
#                else:
 #                   code = esl_codes[ index ]
  #                  if code['code'] != 0:
   #                     print( '\tesl-reformat exit ' + str(code['code'], flush = True) + ' ' + \
    #                        os.path.basename( code['stdin'][-1] ) )
     #               os.remove( code['stdin'][-1] )
        
        print( '\nTrimming alignments' , flush = True)
        trimmed = None
        if os.path.isdir(trm_dir):
            trimmed = collect_files(trm_dir, 'phylip')
        else:
            os.mkdir(trm_dir)
        trim_tuples = compile_trim_cmd(output, mod = trim_mod, trimmed = trimmed)

        trim_codes = multisub(trim_tuples, processes = cpu - 1,
                              verbose = verbose)
        # clipkit or the subsequent command raises an exit error
#        for code in trim_codes:
 #           if code['code']:
  #              eprint('\tERROR: ' + str(code['stdin']), flush = True)



if __name__ == '__main__':

    mp.set_start_method('spawn')
    parser = argparse.ArgumentParser( description = 'Runs `hmmsearch` v each proteome in a' \
        + ' `.db`. For each query, extracts results and optionally outputs a compiled fasta' \
        + ', hmmalignment and/or trimmed alignment.' )

    i_arg = parser.add_argument_group('Inputs')
    i_arg.add_argument('-d', '--mtdb', default = masterDB())
    i_arg.add_argument('--hmm', required = True, help = '.hmm database' )
#    parser.add_argument('-p', '--previous', help = 'Previous db2hmmsearch dir. ' + \
 #       'Be wary of incomplete outputs from interrupted runs' )
#    parser.add_argument('-f', '--fasta', action = 'store_true', \
 #       help = 'Compile fastas for each query')

    p_arg = parser.add_argument_group('Search parameters')
    p_arg.add_argument('-b', '--binary', default = 'hmmsearch',
                       help = 'Search binary {"hmmsearch", \
                               "nhmmer"} NONFUNCTIONAL')
    p_arg.add_argument('-m', '--max_hits', type = int, help = 'Max hits for each organism')
    p_arg.add_argument('-q', '--query_thresh', type = float,
        help = 'Query percent hit threshold (+/-)')
    p_arg.add_argument('-e', '--evalue', help = 'E value threshold, e.g. ' \
        + '10^(-x) where x is the input', type = int)
    p_arg.add_argument('-a', '--accession', action = 'store_true', default = False, \
        help = 'Extract accessions instead of queries (Pfam, etc)' )
    p_arg.add_argument('-l', '--align', default = False, action = 'store_true', \
        help = 'Align fastas to original hmm and trim via clipkit' )

    r_arg = parser.add_argument_group('Runtime options')
    r_arg.add_argument('-v', '--verbose', action = 'store_true')
    r_arg.add_argument('-o', '--output')
    r_arg.add_argument('-c', '--cpu', type = int)

    args = parser.parse_args()
    deps = ['hmmsearch']
    if args.align:
        deps.extend(['hmmalign', 'clipkit', 'esl-reformat'])
    findExecs(deps, exit = set(deps))
#    if args.previous:
 #       args.output = args.previous

    if not args.output:
        output = mkOutput(os.getcwd() + '/', 'db2hmmer')
    else:
        output = format_path(args.output)

    if args.cpu and args.cpu < mp.cpu_count():
        cpu = args.cpu
    else:
        cpu = mp.cpu_count()

    args_dict = { 
        'MycotoolsDB': args.mtdb, 'Hmm database': args.hmm, 'Output': output,
        'Align hmms': args.align, 'Use accession': args.accession, 'Hits/ome': args.max_hits, 
        'Coverage threshold': args.query_thresh, 'E-value': args.evalue, 
        'Output': output, 'CPUs': cpu
        }
    start_time = intro('db2hmmer', args_dict)
    db = mtdb(args.mtdb)

    main(db, format_path(args.hmm), output, 
         args.accession, args.max_hits, args.query_thresh,
         evalue = args.evalue, binary = args.binary, verbose = args.verbose)

    outro(start_time)



#! /usr/bin/env python3

# NEED to implement fa2clus?
# NEED to ignore non fasta inputs

import os
import re
import sys
import argparse
import subprocess
from collections import defaultdict
from mycotools.lib.kontools import eprint, vprint, collect_files, \
    format_path, intro, outro, findExecs, mkOutput, multisub
from mycotools.lib.biotools import fa2dict, dict2fa

class PhyloError(Exception):
    pass

def mafftRun(name, fasta, out_dir, hpc, verbose = True, cpus = 1, spacer = '\t'):

    cmd = [
        'mafft', '--auto', '--thread', '-' + str(cpus), fasta
        ]

    if not hpc:
        print(spacer + 'Aligning', flush = True)
        with open(out_dir + name + '.mafft', 'w') as out_file:
            if verbose:
                run_mafft = subprocess.call(cmd, stdout = out_file)
            else:
                run_mafft = subprocess.call(cmd, stdout = out_file, stderr = subprocess.DEVNULL)
         
        if run_mafft != 0:
            eprint(spacer + '\tERROR: mafft failed: ' + str(run_mafft), flush = True)
            if os.path.isfile(out_dir + name + '.mafft'):
                os.remove(out_dir + name + '.mafft')	
            raise PhyloError
    else:
        with open(out_dir + name + '_mafft.sh', 'w') as out:
            if '#PBS' in hpc:
                out.write( hpc + '\n\n' + ' '.join([str(x) for x in cmd]) + \
                ' > ' + out_dir + '/' + name + '.mafft' + \
                '\ncd ' + out_dir + '\nqsub ' + name + '_clipkit.sh')
            else:
                 out.write( hpc + '\n\n' + ' '.join([str(x) for x in cmd]) + \
		' > ' + out_dir + '/' + name + '.mafft' + \
                '\ncd ' + out_dir + '\nsbatch ' + name + '_clipkit.sh')
        
    return out_dir + '/' + name + '.mafft'


def trimRun(name, mafft, out_dir, hpc, verbose, cpus = 1, spacer = '\t'):

    name2 = out_dir + os.path.basename(mafft) + '.clipkit'
    cmd = ['clipkit', mafft, '--output', name2]
    if not hpc:

        print(spacer + 'Trimming', flush = True)
        if verbose:
            run_clipkit = subprocess.call(cmd, stdout = subprocess.PIPE)
        else:
            run_clipkit = subprocess.call( 
                cmd, stdout = subprocess.DEVNULL, stderr = subprocess.DEVNULL
                )
        if run_clipkit != 0:
            eprint(spacer + '\tERROR: `clipkit` failed: ' + str(run_clipkit) , flush = True)  
            if os.path.isfile(out_dir + name2):
                os.remove(out_dir + name2)
            raise PhyloError

    else:
        with open(out_dir + name + '_clipkit.sh', 'w') as out:
            if '#PBS' in hpc:
                out.write( hpc + '\n\n' + #cmd1 + '\n' + 
                    ' '.join([str(x) for x in cmd]) + '\n\ncd ' + out_dir \
                    + '\nqsub ' + name + '_tree.sh' )
            else:
                out.write( hpc + '\n\n' + #cmd1 + '\n' + 
                    ' '.join([str(x) for x in cmd]) + '\n\ncd ' + out_dir \
                    + '\nsbatch ' + name + '_tree.sh' )
    
    return name2


def run_mf(clipkit_files, out_dir, hpc, verbose, cpus = 1, spacer = '\t'):
    cpus_per_cmd = 3
    concurrent_cmds = round((cpus - 1)/4 - 0.5) # round down
    cmds = [('iqtree', '-m', 'MFP+MERGE', '--threads-max', 
             str(cpus_per_cmd), '-B', '1000', '-s', f_,) \
            for f_ in clipkit_files]
    multisub(cmds, verbose = verbose, processes = concurrent_cmds)

    models = {}
    for f_ in clipkit_files:
        with open(f_ + '.log', 'r') as raw:
            for line in raw:
                if line.startswith('Best-fit model:'):
                    model = re.search(r'Best-fit model: ([^\W]+)', line)[1]
                    models[f_] = model
                    break

    return models

def prepare_nexus(concat_fa, models, spacer = '\t'):

    nex_data = '#nexus\nbegin sets;'
    index = 1
    for i, trimmed_f in enumerate(list(models.keys())):
        trim_fa = fa2dict(trimmed_f)
        len0 = len(trim_fa[list(trim_fa.keys())[0]]['sequence'])
        if not all(len(x['sequence']) == len0 for x in trim_fa.values()):
            eprint(spacer + '\tERROR: alignment sequences are not same length ' \
                   + trimmed_f, flush = True)
            sys.exit(17)
        new_index = index + len0
        nex_data += '\n    charset part' + str(i + 1) + ' = ' \
                  + str(index) + '-' + str(new_index - 1) + ';'
        index = new_index

        for seq, data in trim_fa.items():
            ome = seq[:seq.find('_')]
            if ome in concat_fa:
                concat_fa[ome]['sequence'] += data['sequence']

    nex_data += '\n    charpartition mine = '
    for i, model in enumerate(models.values()):
        nex_data += model + ':part' + str(i+1) + ','
    nex_data = nex_data[:-1] + ';\nend;'

    return nex_data, concat_fa 

def run_partition_tree(fa_file, nex_file, verbose, cpus, spacer = '\t'):
    cmd = ['iqtree', '-s', fa_file, '-p', nex_file, '-T', str(cpus),
           '-B', '1000', '--sampling', 'GENESITE']
    print(spacer + 'Tree building', flush = True)
    if verbose:
        run_tree = subprocess.call(cmd)
    else:
        run_tree = subprocess.call(cmd, stdout = subprocess.DEVNULL,
                                   stderr = subprocess.DEVNULL)
    return run_tree



def treeRun(name, clipkit_file, out_dir, hpc, verbose, fast = False, cpus = 1, spacer = '\t'):

    if fast:
        cmd = ['fasttree', '-out', clipkit_file + '.treefile', clipkit_file]
    else:
        cmd = ['iqtree', '-s', clipkit_file, '-B', '1000', '-T', str(cpus)]
   
    if not hpc:
        print(spacer + 'Tree building', flush = True)
        if verbose:
            run_tree = subprocess.call(
                cmd
                )
        else:
            run_tree = subprocess.call(
                cmd, stdout = subprocess.DEVNULL, stderr = subprocess.DEVNULL
                )
        if run_tree != 0:
            eprint(spacer + '\tERROR: tree failed: ' + str(run_tree), flush = True)
            raise PhyloError
    else: 
        vprint('\nOutputting bash script `tree.sh`.\n', v = verbose, flush = True)

        with open(out_dir + name + '_tree.sh', 'w') as out:
            out.write(hpc + '\n\n' + ' '.join([str(x) for x in cmd]))


def main( 
    fasta_path, slurm = False, torque = False, fast = False, project = '', 
    output_dir = None, verbose = True, alignment = False, mem = '60GB', cpus = 1,
    spacer = '\t\t', align_stop = False, tree_stop = False, 
    flag_incomplete = True, start = 0, partition = False
    ):


    hpc = False
    if slurm:
        vprint('\nHPC mode, preparing submission scripts (Slurm)\n', v = verbose, flush = True)
        hpc = '#!/bin/bash\n#SBATCH --time=24:00:00\n#SBATCH --nodes=1\n' + \
            '#SBATCH --ntasks-per-node=' + str(cpus) + '\n#SBATCH -A ' + project + '\n' + \
            '#SBATCH --mem=' + str(mem)
#            '\n\nsource activate ' + source
    elif torque:
        vprint('\nHPC mode, preparing submission scripts (PBS)\n', v = verbose, flush = True)
        hpc = '#PBS -l walltime=10:00:00\n#PBS -l nodes=1:ppn=4\n#PBS -A ' + project

    output_dir_prep = os.getcwd() + '/'
    if isinstance(fasta_path, str):
        if os.path.isfile(fasta_path):
            if not output_dir:
                dir_name = re.sub(r'\..*?$', '_tree', os.path.basename( os.path.abspath(fasta_path)))
                out_dir = output_dir + '/' + dir_name
                if not os.path.isdir(out_dir):
                    os.mkdir(out_dir)
                out_dir = format_path(out_dir)
            else:
                if not os.path.isdir(output_dir):
                    os.mkdir(output_dir)
                out_dir = format_path(output_dir)
            files = [fasta_path]
        elif os.path.isdir(fasta_path):
            if not output_dir:
                out_dir = mkOutput(output_dir_prep, 'fa2tree')
            else:
                if not os.path.isdir(output_dir):
                    os.mkdir(output_dir)
                out_dir = format_path(output_dir)
            files = collect_files(fasta_path, '*')
            check_fas = set(files)
            new_fas = []
            for f in files:
                if f + '.mafft.clipkit' not in check_fas:
                    if f + '.mafft' not in check_fas:
                        new_fas.append(f)
            files = new_fas
    else:
        if not output_dir:
            out_dir = mkOutput(output_dir_prep, 'fa2tree')
        else:
            if not os.path.isdir(output_dir):
                os.mkdir(output_dir)
            out_dir = format_path(output_dir)
        files = []
        for path in fasta_path:
            if os.path.isdir(path):
                files.extend([format_path(x) for x in collect_files(path, '*')])
            else:
                files.append(path)

    # check for non fasta inputs
    if any(f_.endswith(('.nexus', '.nex', '.nxs', 
                        '.phylip', '.phy', '.ph',
                        '.clustal', '.clus',)) \
           for f_ in files):
        print(spacer + 'Converting to fastas', flush = True)
        from Bio import SeqIO
        conv_dir = out_dir + 'conv/'
        if not os.path.isdir(conv_dir):
            os.mkdir(conv_dir)
        for i, f_ in enumerate(files):
            out_name = conv_dir + os.path.basename(
                re.search(r'(.*)\.[^.]+$', f_)[1] + '.fa'
                )
            if not os.path.isfile(out_name):
                if f_.endswith(('.nexus', '.nex', '.nxs',)):
                    records = SeqIO.parse(f_, 'nexus')
                    SeqIO.write(records, out_name, 'fasta')
                elif f_.endswith(('.phylip', '.phy', '.ph',)):
                    records = SeqIO.parse(f_, 'phylip')
                    SeqIO.write(records, out_name, 'fasta')
                elif f_.endswith(('.clustal', '.clus',)):
                    records = SeqIO.parse(f_, 'clustal')
                    SeqIO.write(records, out_name, 'fasta')
                else:
                    continue
            files[i] = out_name

    if partition: #multigene partition mode
        ome2fa2gene = defaultdict(dict)
        incomp_omes = {}
        for f in files:
            if f.endswith('/'):
                f = f[:-1]
            fa_dict = fa2dict(f)
            incomp_omes[os.path.basename(f)] = []
            for seq in fa_dict:
                ome = seq[:seq.find('_')]
                if f not in ome2fa2gene[ome]:
                    incomp_omes[os.path.basename(f)].append(ome)
                    ome2fa2gene[ome][f] = seq
                else:
                    eprint('\nERROR: multiple sequences for a ' \
                         + 'single ome in ' + f, flush = True)
                    sys.exit(5)

        incomp_files, comp_files = {}, []
        for f, omes in incomp_omes.items():
            if len(omes) < len(ome2fa2gene):
                incomp_files[f] = sorted(
                    set(ome2fa2gene.keys()).difference(set(omes))
                    )
            else:
                comp_files.append(f)

        if incomp_files:
            incomp_files = {k: v for k,v in sorted(incomp_files.items(), 
                                                   key = lambda x: len(x[1]))}
            if flag_incomplete:
                eprint('\nERROR: omes without sequences in all fastas: ', flush = True)
            else:
                eprint('\nWARNING: omes removed without sequences in all \
                        fastas: ', flush = True)
            for f, omes in incomp_files.items():
                eprint('\t' + f + ': ' + ','.join([x for x in omes]), flush = True)
            if comp_files:
                eprint('\tComplete files: ' + ','.join(comp_files), flush = True)
            else:
                eprint('\tNo complete files', flush = True)
            if flag_incomplete:
                sys.exit(6)


    trimmed_files = []
    for fasta in files:
        name = re.search(r'.*?/*([^/]+)/*$', fasta)[1]
        fasta_name = os.path.basename(os.path.abspath(fasta))
        clipkit = out_dir + fasta_name + '.clipkit'
        if start == 0:
            mafft = out_dir + fasta_name + '.mafft'
            clipkit = mafft + '.clipkit'
            try:
                if os.stat(mafft):
                    vprint('\nAlignment exists', v = verbose, flush = True)
                else:
                    raise ValueError
            except (FileNotFoundError, ValueError) as e:
                if not alignment:
                    mafft = mafftRun(name, os.path.abspath(fasta), out_dir, hpc, 
                                     verbose, cpus, spacer = spacer)
                else:
                    mafft = fasta
        else:
            mafft = fasta
            clipkit = mafft + '.clipkit'


        if start == 0 or start == 1: 
            try:
                if os.stat(clipkit):
                    vprint('\nTrim exists', v = verbose, flush = True)
                    clipkit_out = clipkit
                else:
                    raise ValueError
            except (FileNotFoundError, ValueError) as e:
                clipkit_out = trimRun(name, mafft, out_dir, hpc, verbose, spacer = spacer)
            trimmed_files.append(clipkit_out)
        else:
            trimmed_files.append(fasta)

    if align_stop:
        print(spacer + 'Alignments outputed', flush = True)
        sys.exit(0)

    if not partition:
        for trimmed_f in trimmed_files:
            name_prep = re.search(r'.*?/*([^/]+)/*$', trimmed_f)[1]
            name = re.sub(r'\.mafft\.clipkit$', '', name_prep)
            treeRun(name, trimmed_f, out_dir, hpc, verbose, fast, cpus, spacer = spacer)
    else:
        print(spacer + 'Model finding', flush = True)
        models = run_mf(trimmed_files, out_dir, hpc, verbose, cpus,
               spacer = spacer + '\t')

        print(spacer + 'Concatenating', flush = True)
        concat_fa = {ome: {'description': '', 'sequence': ''}
                     for ome in ome2fa2gene if ome not in incomp_omes}
        nex_data, concat_fa = prepare_nexus(concat_fa, models)
        with open(out_dir + 'concatenated.fa', 'w') as out:
            out.write(dict2fa(concat_fa))
        with open(out_dir + 'concatenated.nex', 'w') as out:
            out.write(nex_data)

        if tree_stop:
            print(spacer + 'Concatenated nexus and fasta outputted', 
                  flush = True)
            sys.exit(0)

        run_partition_tree(out_dir + 'concatenated.fa', 
                           out_dir + 'concatenated.nex',
                           verbose, cpus, spacer = spacer)

    if hpc:
        if slurm:
            print('\nStart pipeline via `sbatch <STARTSTEP>.sh` in ' \
                + out_dir + '\n')
        else:
            vprint('\nStart pipeline via `qsub <STARTSTEP>.sh` in ' + out_dir + '\n',
                v = verbose)


if __name__ == '__main__':

    parser = argparse.ArgumentParser( 
        description = 'Takes in a multifasta, or directory/list of fastas [partition]  ' + \
        ' aligns (mafft), trims (clipkit), and infers phylogeny \
        (iqtree/fastree). When using non-Mycotools fastas for multigene \
        partition analysis, all accessions must be formatted as \
        "<GENOME>_<ACCESSION>" and each <GENOME> needs to be in ALL fastas'
        )

    io_opt = parser.add_argument_group('Input options')
    io_opt.add_argument('-i', '--input', required = True,
        help = 'Fasta or directory of fastas/alignments (fasta, phylip, \
                nexus, or clustal). "-" for stdin')
    io_opt.add_argument('-a', '--align', help = 'Start from alignments',
                        action = 'store_true')
    io_opt.add_argument('-t', '--trim', help = 'Start from trimmed alignments',
                        action = 'store_true')

    s_opt = parser.add_argument_group('Single tree options')
    s_opt.add_argument('--fast', action = 'store_true', \
        help = 'Fasttree' )
    s_opt.add_argument('-s', '--slurm', action = 'store_true', \
        help = 'Submit ALL steps to Slurm' )
    s_opt.add_argument('--torque', action = 'store_true', \
        help = 'Submit ALL steps to Torque.' )
    s_opt.add_argument('-A', '--project', help = 'HPC project')

    m_opt = parser.add_argument_group('Partition tree options')
    m_opt.add_argument('-p', '--partition', action = 'store_true',
        help = 'Multigene partition mode')
    m_opt.add_argument('-sa', '--stop_align', action = 'store_true',
        help = 'Stop after aligning and trimming')
    m_opt.add_argument('-sc', '--stop_cat', action = 'store_true',
        help = 'Stop after concatenating sequences')
    m_opt.add_argument('-m', '--missing', action = 'store_true',
        help = 'Remove omes with missing sequences')

    r_opt = parser.add_argument_group('Runtime options')
    r_opt.add_argument('-v', '--verbose', action = 'store_true')
    r_opt.add_argument('-o', '--output')
    r_opt.add_argument('-c', '--cpus', default = 1, type = int)

    args = parser.parse_args()

    execs = ['mafft', 'clipkit']
    if args.fast:
        execs.append('fasttree')
    else:
        execs.append('iqtree')
    if args.torque or args.slurm:
        findExecs(execs, execs)
    else:
        findExecs(execs, execs)

    output = format_path(args.output)
    if output:
        if not output.endswith('/'):
            output += '/' # bring to mycotools path expectations
    args_dict = {
        'Fasta': args.input, 'Fast': args.fast, 
        'Multigene partition': args.partition, 'Input alignments': bool(args.align),
        'Input trimmed': bool(args.trim),
        'Alignment stop': args.stop_align, 'Concatenation stop': args.stop_cat,
        'Ignore missing': args.missing, 'Torque': args.torque, 'Slurm': args.slurm, 
        'HPC project': args.project, 'Output': output, 'CPUs': args.cpus
        }
    start_time = intro('Fasta2Tree', args_dict)


    if args.input == '-':
        args.input = \
            stdin2str().replace('"','').replace("'",'').split()
    else:
        args.input = args.input.replace('"','').replace("'",'').split()


    start = 0
    if args.trim:
        start = 2
    elif args.align:
        start = 1
    main( 
        args.input, slurm = args.slurm, fast = args.fast,
        torque = args.torque, project = args.project,
        output_dir = output, verbose = bool(args.verbose), 
        alignment = args.align,
        align_stop = args.stop_align, tree_stop = args.stop_cat,
        flag_incomplete = bool(not args.missing), start = start,
        partition = bool(args.partition), cpus = args.cpus
        )
#    elif os.path.isdir(input_check):
 #       fas = collect_files(input_check, 'fa')
  #      fas.extend(collect_files( input_check, 'fasta' ))
#        fas.extend(collect_files( input_check, 'fna' ))
 #       fas.extend(collect_files( input_check, 'fsa' ))
  #      if not fas:
   #         eprint('\nERROR: no {.fa, .fasta, .fna, .fsa} detected', flush = True)
    #        sys.exit(2)
     #   print(flush = True)
      #  for fa in fas:
       #     print(fa, flush = True)
        #    main(
         #       fa, slurm = args.slurm, torque = args.torque, alignment = alignment, fast = args.fast,
          #      project = args.project, output_dir = output, verbose = False, spacer = '\n'
           #     )
#    else:
 #       eprint('\nERROR: invalid input', flush = True)
  #      sys.exit(3)

    outro(start_time)



#! /usr/bin/env python3

import os
import re
import sys
import argparse
from mycotools.lib.kontools import intro, outro, file2list, format_path, mkOutput

def grab_names(data, query = False):

    if query:
        comp = re.compile(r'Query\:       (.*?)\W+\[M\=\d+\]')
    else:
        comp = re.compile(r'Accession\:   (.*)')

    namesPrep = comp.finditer(data)
#    namesPrep = comp.findall(data)
    names_set = set(x[1] for x in namesPrep)

    return names_set


def grab_hits(data, threshold = None, evalue = None, 
             top = None, query = False, accession = False):

    if query:
        comp = re.compile( r'^Query: +' + query + r' +\[M\=(\d+)\]$\n([\s\S]+?\n\n)' \
            + r'([\s\S]*?\n\n\n)', re.M )
    else:
        comp = re.compile( r'Query\:.*?\[M\=(\d+)\]$\n(^Accession\: +' + accession \
            + r'[\s\S]+?\n\n)([\s\S]*?\n\n\n)', re.M )

    hitsAligns = comp.search( data )
    if hitsAligns is not None:
        querySize = int(hitsAligns[1])
        fullHit = hitsAligns[2]
        fullAlign = hitsAligns[3]
        lines = fullHit.split('\n')
        prepLines = []
        for line in lines:
            if re.search(r'^\W+\d', line):
                prepLines.append(line)

        outputLines = []
        for line in prepLines:
            outPrep = line.split(' ')
            outLine = [x for x in outPrep if x != '']
            outputLines.append( outLine )

        hit_str = ''
        for line in outputLines:
            hit_str += ' '.join(line[8:]) + '\t' + line[0] + '\t' + line[1] + '\t' + line[2] + '\t' + \
                line[3] + '\t' + line[4] + '\t' + line[5] + '\t' + line[6] + '\t' + line[7] + '\n'

        passingSeq, align_str, count = set(), '', 0
        aligns = re.findall( r'^\>\> (.*?)\n.*?\n[\W\-]+\n(.*?)\n\n', fullAlign, re.DOTALL | re.MULTILINE )
        for align in aligns:
            seq = align[0].rstrip()
            lines = align[1].split('\n')
            for line in lines:
                outPrep = line.split(' ')
                outLine = [x for x in outPrep if x != '' and x != '..' and '[' not in x and ']' not in x]
                if len(outLine) != 13:
                    print('\nINVALID ALIGNMENT HEADERS - CHECK SCRIPT/ALIGNMENT\n', flush = True)
                    sys.exit( 5 )
                if threshold:
                    hmmStart = int(outLine[6])
                    hmmEnd = int(outLine[7])
                    hitSize = hmmEnd - hmmStart + 1
                    if hitSize < threshold * querySize:
                        continue
                    passingSeq.add( str(seq) )
                    

                align_str += str(seq) + '\t' + outLine[0] + '\t' + outLine[1] + '\t' + outLine[2] + \
                    '\t' + outLine[3] + '\t' + outLine[4] + '\t' + outLine[5] + '\t' + outLine[6] + '\t' + \
                    outLine[7] + '\t' + outLine[8] + '\t' + outLine[9] + '\t' + outLine[10] + '\t' + \
                    outLine[11] + '\t' + outLine[12] + '\n'

        if threshold:
            hit_list = [ hit for hit in hit_str.split( '\n' ) if hit != '' ]
            out_hits = []
            for hit in hit_list:
                start = re.search(r'^(.*?)\t', hit)[1]
                if start.rstrip() in passingSeq:
                    out_hits.append( hit )
            hit_str = '\n'.join( out_hits )

        if evalue:
            new_passSeq = set()
            hit_list = [x for x in hit_str.split('\n') if x]
            align_list = [x for x in align_str.split('\n') if x]
            out_hits, out_aligns = [], []
            for hit in hit_list:
                if int(10**200 * float( hit.split('\t')[1] )) < int(10**200 * float( '1e-' + str(evalue) )):
                    out_hits.append( hit )
                    new_passSeq.add( hit.split('\t')[0] )
            for align in align_list:
                if align.split('\t')[0] in new_passSeq:
                    out_aligns.append( align )
            hit_str = '\n'.join( out_hits )
            align_Str = '\n'.join( out_aligns )

        if top:
            count = 0
            hit_list, align_list = hit_str.split( '\n' ), align_str.split( '\n' )
            out_hits, out_aligns = [], []
            for hit in hit_list:
                count += 1
                if count > top:
                    break
                out_hits.append( hit )
            count = 0
            for align in align_list:
                count += 1
                if count > top:
                    break
                out_aligns.append( align )
            hit_str = '\n'.join( out_hits )
            align_str = '\n'.join( out_aligns )

    else:
        hit_str, align_str = None, None
        print( 'did not work' , flush = True)

    return hit_str, align_str


def synthesizeHits( out_dict ):

    check = []
    for hit in out_dict:
        names = set( x.split('\t')[0] for x in out_dict[ hit ][0].split('\n') )
        check.append( names )
    
    passing = set()
    for names in check:
        for name in list(names):
            if all( name in x for x in check ):
                passing.add( name )
    
    new_dict = {}
    for hit in out_dict:
        hit_str, align_str = '', ''
        for seq in out_dict[ hit ][ 0 ].split('\n'):
            if seq.split('\t')[0] in passing:
                hit_str += seq + '\n'
        for seq in out_dict[ hit ][ 1 ].split( '\n' ):
            if seq.split('\t')[0] in passing:
                align_str += seq + '\n'
        new_dict[ hit ] = ( hit_str, align_str )

    return new_dict

    
def main( 
    data, accession, best, threshold, evalue,
    query = True, header = True 
    ):

    check, out = None, {}
    if accession:
#        if accession == True:
        check = grab_names(data)
        if not check:
            check = grab_names(data, True)
            accession = False
            query = True
    else:
#        if query == True:
        check = grab_names(data, True)
        if not check:
            accession = True
            query = False
            check = grab_names(data)

    if isinstance(check, set):
        for x in check:
            hits, align = '', ''
            if header: 
                hits = '#seq\tseq_e\tseq_score\tseq_bias\tdom_e\tdom_score\tdom_bias\texp\tN\n'
                align = '#seq\tdomNum\tincl_thresh\tscore\tbias\tc-Evalue\ti-Evalue\thmmFrom\t' + \
                                    'hmmTo\taliFrom\taliTo\tenvFrom\tenvTo\tacc\n'
            if accession:
                t_hits, t_aligns = grab_hits( data, threshold, evalue = evalue, \
                    top = best, accession = x )
            else:
                t_hits, t_aligns = grab_hits( data, threshold, evalue = evalue, \
                    top = best, query = x )
            if t_hits:
                hits += t_hits
                align += t_aligns
                out[x] = ( hits, align )
    else:
        htis, align = '', ''
        if header:
            hits = '#seq\tseq_e\tseq_score\tseq_bias\tdom_e\tdom_score\tdom_bias\texp\tN\n'
            align = '#seq\tdomNum\tincl_thresh\tscore\tbias\tc-Evalue\ti-Evalue\thmmFrom\t' + \
                            'hmmTo\taliFrom\taliTo\tenvFrom\tenvTo\tacc\n'
        t_hits, t_aligns = grab_hits(data, threshold, top = best, \
            query = query, accession = accession)
        hits += t_hits
        align += t_aligns
        if accession:
            out[ accession ] = ( hits, align )
        else:
            out[ query ] = ( hits, align )

    return out
            

if __name__ == '__main__':

    parser = argparse.ArgumentParser( description = 'Extracts inputted accession hits and alignments from `hmmsearch` ' + \
        'output. Optionally provide a size-based query threshold to only extract alignments of a certain size.' )
    parser.add_argument( '-i', '--input', required = True, help = '`hmmsearch` output' )
    parser.add_argument( '-a', '--accession', \
        help = 'Accession to extract or new line delimited list of accessions; `-a`|`-q` required' )
    parser.add_argument( '-q', '--query', \
        help = 'Query to extract or new line delimited list of queries; `-a`|`-q` required' )
    parser.add_argument( '-b', '--best', type = int, help = '# top hits to take' )
    parser.add_argument( '-t', '--threshold', type = float, help = 'Percent query coverage to consider a hit (float).' )
    parser.add_argument( '-e', '--evalue', help = 'Negative magnitude of E-value, e.g. ' \
        + '10^(-x) where `x` is the input.' )
    parser.add_argument( '--allq', action = 'store_true', help = 'Extract all queries' )
    parser.add_argument( '--alla', action = 'store_true', help = 'Extract all accessions' )
    parser.add_argument( '-o', '--output', help = 'Output file name/path (extensions automatically applied' )
    args = parser.parse_args()

    # initialize output file structure
    output = format_path(args.output)
    if not args.output:
        output = mkOutput(os.getcwd() + '/', 'extractHmmsearch')
    elif not os.path.isdir(output):
        os.mkdir(args.output)

    args_dict = { 
        'Input': args.input, 'Output': output, 'Accession': args.accession,
        'Query': args.query,  'Best hits': args.best, 'Threshold': args.threshold,
        'E-value': args.evalue
    }
    start_time = intro( 'hmmsearch Extraction', args_dict )


    if args.allq:
        args.query = True
    elif args.alla:
        args.accession = True
    else:
        if args.query:
            if os.path.isfile( args.query ):
                ques = file2list( args.query )
            else:
                ques = [ args.query ]
            args.query = True
        elif args.accession:
            if os.path.isfile( args.accession ):
                ques = file2list( args.accession )
            else:
                ques = [ args.accession ]
            args.accession = True
        else:
            print('\nNeed `-q` or `-a` specified\n', flush = True)
            sys.exit(8)
        

    if not os.path.isfile( args.input ):
        print('\n\tNot a valid input file\n', flush = True)
        sys.eixt(2)
    with open( args.input, 'r' ) as raw:
        data = raw.read()


    if args.allq or args.alla:
        out_dict = main( 
            data, args.accession, args.best, args.threshold, args.evalue, query = args.query 
            )
    else:
        for que in ques:
            if args.query:
                args.query = que
            else:
                args.accession = que   
            temp_out_dict = main( 
                data, args.accession, args.best, args.threshold, 
                args.evalue, header = False, query = args.query 
                )
            out_dict = { **out_dict, **temp_out_dict }
        out_dict = synthesizeHits( out_dict )

    for name in out_dict:
        with open(output + '/' + name + '.hits.tsv', 'w') as out:
            out.write( out_dict[name][0] )
        with open(output + '/' + name + '.aligns.tsv', 'w') as out:
            out.write( out_dict[name][1] )

    outro( start_time )
